//Includes
#include <Arduino.h>
#include "ADS131M04.h"
#include <Wire.h>
#include <MCP23017.h>
//#include <MAX31865.h>
#include <SPI.h>
#include "FS.h"
#include "SD.h"
#include <Preferences.h>

//Constants
#define MCP23017_ADDR 0x20

#define MCP_INT_A 25
#define MCP_INT_B 26
#define SD_CD 2
#define ADC_NDRDY 16
#define ANALOG_VCC 36
#define MAX1_NDRDY 35
#define MAX2_NDRDY 34

#define RESET_PERIF 4
#define SD_NCS 5
#define ADC_NCS 15
#define MAX1_CS 32
#define MAX2_CS 33

#define RREF      390.0
#define RNOMINAL  100.0

//Variables saved also in Flash memory
boolean loggSD = 1;
boolean loggPC = 0;
double RTDa = 3.9083e-3;
double RTDb = -5.775e-7;
int periodSD = 500;
int periodPC = 500;
String filePathName = "/data.csv";

double U1_G_C = 0.9883266733;
double U1_O_C = -0.00076;
double I1_UG_C = 0.9391435;
double I1_UO_C = 0.0065;
double I1_LG_C = 0.98745927;
double I1_LO_C = 0.0006307;
double U2_G_C = 0.984425362;
double U2_O_C = -0.00185;
double I2_UG_C = 0.93931993;
double I2_UO_C = 0.0065;
double I2_LG_C = 0.98463962;
double I2_LO_C = 0.0006255;

//Variables
double resistance;
double temperature;
uint8_t reg1, reg2; //reg1 = MSB, reg2 = LSB
uint16_t fullreg; //= MSB + LSB
double temp;
volatile bool interrupted = false;
boolean range_1 = 0;
boolean range_2 = 0;
unsigned long warningTime = 0;
unsigned long dataOrder = 0;
unsigned long flashTime = 3;
int delay1 = 1;
int delay2 = 1;
int blueTime = 25;
int result = 0;
double measuredDataADC[] = {0.0, 0.0, 0.0, 0.0};
double all_data[] = {0.0, 0.0, 0.0, 0.0, 0.0, 0.0};
double Z1, Z2, Z3, Z4, Rt;
double rpoly = 0;
double sampleID = 0;
String dataMessage;
double timeSD = 0;
double timePC = 0;
String dataToPrint = "";
int numOfSamples = 5;
int count_r1 = 0;
int count_r2 = 0;
String timeStamp = "";
String inData[4];
const String defFPN = "/data.csv";
unsigned long checkTime = 0;

//Objects
ADS131M04 adc;
MCP23017 mcp = MCP23017(MCP23017_ADDR);
SPIClass spiSD = SPIClass(HSPI);
Preferences preferences;
//----------------------------------------------------------------------------
void userBut() {
  interrupted = true;
}
//----------------------------------------------------------------------------
void setup() {
  Serial.begin(9600); //baudrate
  Wire.begin();
  SPI.begin(14, 12, 13); //SCK, MISO, MOSI
  adc.begin(14, 12, 13, ADC_NCS, ADC_NDRDY); //SCK, MISO, MOSI, CS, DRDY

  Serial.println();

  /*
    preferences.begin("card", false);
    preferences.clear();
    preferences.putBool("loggSD", 0);
    preferences.putBool("loggPC", 0);
    preferences.putDouble("RTDa", RTDa);
    preferences.putDouble("RTDb", RTDb);
    preferences.putInt("periodSD", 1000);
    preferences.putInt("periodPC", 1000);
    //preferences.putString("filePathName", filePathName);
    preferences.end();
  */

  preferences.begin("card", false);
  loggSD = preferences.getBool("loggSD", -1);
  loggPC = preferences.getBool("loggPC", -1);
  RTDa = preferences.getDouble("RTDa", -1);
  RTDb = preferences.getDouble("RTDb", -1);
  periodSD = preferences.getInt("periodSD", -1);
  periodPC = preferences.getInt("periodPC", -1);
  //filePathName = preferences.getString("filePathName");
  preferences.end();

  Serial.println(loggSD);
  Serial.println(loggPC);
  Serial.println(RTDa);
  Serial.println(RTDb);
  Serial.println(periodSD);
  Serial.println(periodPC);
  Serial.println(filePathName);

  pinMode(RESET_PERIF, OUTPUT);
  pinMode(SD_NCS, OUTPUT);
  pinMode(ADC_NCS, OUTPUT);
  pinMode(MAX1_CS, OUTPUT);
  pinMode(MAX2_CS, OUTPUT);

  pinMode(MCP_INT_A, INPUT);
  pinMode(MCP_INT_B, INPUT);
  pinMode(SD_CD, INPUT);
  pinMode(ADC_NDRDY, INPUT);
  pinMode(MAX2_NDRDY, INPUT);
  pinMode(MAX1_NDRDY, INPUT);

  digitalWrite(RESET_PERIF, LOW);
  digitalWrite(SD_NCS, HIGH);
  digitalWrite(ADC_NCS, HIGH);
  digitalWrite(MAX1_CS, HIGH);
  digitalWrite(MAX2_CS, HIGH);

  analogReadResolution(12); //nastaveni ADC u ESP na 12 bitu
  analogSetAttenuation(ADC_0db); //utlum na vstupu ADC u ESP 0 dB

  attachInterrupt(digitalPinToInterrupt(MCP_INT_B), userBut, FALLING);

  digitalWrite(RESET_PERIF, HIGH);
  delay(1);
  digitalWrite(RESET_PERIF, LOW);
  delay(100);

  mcp.init();
  if (digitalRead(SD_CD) == 0)
  {
    spiSD.begin(18, 19, 23); //SCK, MISO, MOSI
    startSD(filePathName);
  }

  setMCP();
  setADC();
  KR();
  //Serial.println("OK");
}
//----------------------------------------------------------------------------
void loop()
{
  if (Serial.available())
  {
    recvBytes();
    //recvCommand();
  }

  if (((millis() - timePC) > periodPC) && (loggPC == 1))
  {
    Serial.println(measALL());
    timePC = millis();
  }

  if (((millis() - timeSD) > periodSD) && (loggSD == 1) && digitalRead(SD_CD) == 0)
  {
    appendFile(SD, filePathName.c_str(), (((String)(dataOrder) + ";" + measALL_raw() + "\n").c_str()));
    dataOrder = dataOrder + 1;
    timeSD = millis();
  }

  checkBut();
  if ((millis() - checkTime) > 250)
  {
    checkRange();
    checkTime = millis();
  }
}
//----------------------------------------------------------------------------
void checkBut()
{
  uint8_t a, b;
  uint8_t firstA, firstB;
  uint8_t secondA, secondB;

  if (!interrupted)
  {
    mcp.clearInterrupts();
    return;
  }
  //debouncing
  delay(100);
  interrupted = false;

  // stav registru v okamžiku vyvolání přerušení
  mcp.interruptedBy(a, b);
  mcp.clearInterrupts(firstA, firstB);
  // stav registru po časovém zpoždění kvůli zákmitům
  secondB = mcp.readPort(MCP23017Port::B);

  // pokud je stejný stav v okamžiku přerušení i po delay
  if ((b & ~secondB) == (b & ~firstB)) {
    secondA = mcp.readPort(MCP23017Port::A);

    // pokud je stisknuto tlačítko BTN_1
    if ((b & 0b00000011) == 0b01)
    {
      //co se ma stat pro stisknuti tohoto tlacitka
      loggSD = !loggSD;
      dataOrder = 0;
      preferences.begin("card", false);
      preferences.putBool("loggSD", loggSD);
      preferences.end();
      //filePathName = defFPN;

      if (loggSD == 1)
      {
        appendFile(SD, filePathName.c_str(), "\n---------- New data package ----------\n\n");
        appendFile(SD, filePathName.c_str(), "Sample no./Timestamp;U_1 (V);I_1 (A);U_2 (V);I_2 (A);T_1 (°C);T_2 (°C) \r\n");
        Serial.print("Writing to: ");
        Serial.println(filePathName);
      }
      else
      {
        filePathName = defFPN;
      }
      delay(550);
    }
    // pokud je stisknuto tlačítko BTN_2
    else if ((b & 0b00000011) == 0b10)
    {
      loggPC = !loggPC;
      preferences.begin("card", false);
      preferences.putBool("loggPC", loggPC);
      preferences.end();
      delay(550);
    }
    else {}
  }
}
//----------------------------------------------------------------------------
void recvCommand()
{
  int i = 0;
  while (Serial.available() > 0)
  {
    char recv = Serial.read();
    if (recv == '\n')
    {
      checkCommand(inData[0], inData[1], inData[2], inData[3]);
      inData[0] = "";
      inData[1] = "";
      i = 0;
    }
    else if (recv == '*')
    {
      i = 1;
    }
    else
    {
      inData[i] += recv;
    }
  }
}
//----------------------------------------------------------------------------
void recvBytes()
{
  int p = 0;
  char buff[100];
  Serial.readBytes(buff, 100);
  for (int i = 0; i < 100; i++)
  {
    if (buff[i] == '|')
    {
      p += 1;
    }
    else if (buff[i] == '*')
    {
      checkCommand(inData[0], inData[1], inData[2], inData[3]);
    }
    else
    {
      inData[p] = inData[p] + buff[i];
    }
    //Serial.println(buff[i]);
  }

  inData[0] = "";
  inData[1] = "";
  inData[2] = "";
  inData[3] = "";

  int l = 0;
  while (l < 4)
  {
    inData[l] = "";
    l++;
  }
}
//----------------------------------------------------------------------------
void checkCommand(String message, String messageVal_1, String messageVal_2, String messageVal_3)
{
  if (message == "OK?")
  {
    Serial.println("OK");
  }
  else if (message == "GA")
  {
    Serial.println(measALL());
  }
  else if (message == "LPCon")
  {
    loggPC = 1;
    preferences.begin("card", false);
    preferences.putBool("loggPC", loggPC);
    //loggPC = preferences.getBool("loggPC", -1);
    preferences.end();
    Serial.println("OK");
  }
  else if (message == "LPCoff")
  {
    loggPC = 0;
    //filePathName = defFPN;
    preferences.begin("card", false);
    preferences.putBool("loggPC", loggPC);
    //loggPC = preferences.getBool("loggPC", -1);
    preferences.end();
    Serial.println("OK");
  }
  else if (message == "LSDon")
  {
    loggSD = 1;
    preferences.begin("card", false);
    preferences.putBool("loggSD", loggSD);
    //loggSD = preferences.getBool("loggSD", -1);
    preferences.end();
    Serial.print("Writing to: ");
    Serial.println(filePathName);
    Serial.println("OK");
  }
  else if (message == "LSDoff")
  {
    loggSD = 0;
    //filePathName = defFPN;
    preferences.begin("card", false);
    preferences.putBool("loggSD", loggSD);
    //loggSD = preferences.getBool("loggSD", -1);
    preferences.end();
    Serial.println("OK");
  }
  else if (message == "FPC")
  {
    periodPC = int(1000 * (1.0 / (messageVal_1.toDouble())));
    preferences.begin("card", false);
    preferences.putInt("periodPC", periodPC);
    //periodPC = preferences.getInt("periodPC", -1);
    preferences.end();
    Serial.println("OK");
  }
  else if (message == "FSD")
  {
    periodSD = int(1000 * (1.0 / (messageVal_2.toDouble())));
    preferences.begin("card", false);
    preferences.putInt("periodSD", periodSD);
    //periodSD = preferences.getInt("periodSD", -1);
    preferences.end();
    Serial.println("OK");
  }
  else if (message == "VCC")
  {
    //Serial.print("VCC voltatage: ");
    Serial.println((1.08 / 4095.0) * analogRead(ANALOG_VCC) * 5.738);
    //Serial.print(analogRead(ANALOG_VCC));
    //Serial.println(" V");
    Serial.println("OK");
  }
  else if (message == "FN")
  {
    if (messageVal_3 != "")
    {
      filePathName = '/' + messageVal_3 + ".csv";
      checkSD(filePathName);
    }
    else
    {
      Serial.println("No set a file name");
    }
    //Serial.print("Set file path: ");
    //Serial.println(filePathName);
    Serial.println("OK");
  }
  else if (message == "PAR")
  {
    periodPC = int(1000 * (1.0 / (messageVal_1.toDouble())));
    preferences.begin("card", false);
    preferences.putInt("periodPC", periodPC);
    //periodPC = preferences.getInt("periodPC", -1);
    preferences.end();
    Serial.println("OK");

    periodSD = int(1000 * (1.0 / (messageVal_2.toDouble())));
    preferences.begin("card", false);
    preferences.putInt("periodSD", periodSD);
    //periodSD = preferences.getInt("periodSD", -1);
    preferences.end();
    Serial.println("OK");

    if (messageVal_3 != "")
    {
      filePathName = '/' + messageVal_3 + ".csv";
      checkSD(filePathName);
    }
    else
    {
      Serial.println("No set a file name");
    }
    //Serial.print("Set file path: ");
    //Serial.println(filePathName);
    Serial.println("OK");
  }
  /*else if ()
    {

    }*/
  else
  {
    Serial.print("No matching message to: ");
    Serial.print(message);
    //Serial.print("");
    Serial.println(messageVal_1 + messageVal_2 + messageVal_3);
    message = "";
    messageVal_1 = "";
    messageVal_2 = "";
    messageVal_3 = "";
  }
}
//----------------------------------------------------------------------------
String measALL()
{
  String meas_values = "";
  double meas_values_u_1[numOfSamples] = {0};
  double meas_values_i_1[numOfSamples] = {0};
  double meas_values_u_2[numOfSamples] = {0};
  double meas_values_i_2[numOfSamples] = {0};

  for (int i = 0; i < numOfSamples; i++)
  {
    readADC();
    meas_values_u_1[i] = (measuredDataADC[0]);
    meas_values_i_1[i] = (measuredDataADC[1]);
    meas_values_u_2[i] = (measuredDataADC[2]);
    meas_values_i_2[i] = (measuredDataADC[3]);
    //delayMicroseconds(10);
  }

  double sum_data[4] = {0.0, 0.0, 0.0, 0.0};

  for (int i = 0; i < numOfSamples; i++)
  {
    sum_data[0] = sum_data[0] + meas_values_u_1[i];
    sum_data[1] = sum_data[1] + meas_values_i_1[i];
    sum_data[2] = sum_data[2] + meas_values_u_2[i];
    sum_data[3] = sum_data[3] + meas_values_i_2[i];
  }

  //readADC();
  meas_values = "data;" +
                (String)((int)(100000 * (sum_data[0] / numOfSamples))) + ';' +
                (String)((int)(100000 * (sum_data[1] / numOfSamples))) + ';' +
                (String)((int)(100000 * (sum_data[2] / numOfSamples))) + ';' +
                (String)((int)(100000 * (sum_data[3] / numOfSamples)));
  readRegRTD(1);
  meas_values = meas_values + ';' + (String)((int)(10 * (convertToTemperature())));
  //LED_ch_x(1);
  readRegRTD(2);
  meas_values = meas_values + ';' + (String)((int)(10 * (convertToTemperature())));
  //LED_ch_x(2);
  //delay(1);
  return (meas_values);
}
//----------------------------------------------------------------------------
String measALL_raw()
{
  String meas_values = "";
  double meas_values_u_1[numOfSamples] = {0};
  double meas_values_i_1[numOfSamples] = {0};
  double meas_values_u_2[numOfSamples] = {0};
  double meas_values_i_2[numOfSamples] = {0};

  for (int i = 0; i < numOfSamples; i++)
  {
    readADC();
    meas_values_u_1[i] = (measuredDataADC[0]);
    meas_values_i_1[i] = (measuredDataADC[1]);
    meas_values_u_2[i] = (measuredDataADC[2]);
    meas_values_i_2[i] = (measuredDataADC[3]);
    //delayMicroseconds(10);
  }

  double sum_data[4] = {0.0, 0.0, 0.0, 0.0};

  for (int i = 0; i < numOfSamples; i++)
  {
    sum_data[0] = sum_data[0] + meas_values_u_1[i];
    sum_data[1] = sum_data[1] + meas_values_i_1[i];
    sum_data[2] = sum_data[2] + meas_values_u_2[i];
    sum_data[3] = sum_data[3] + meas_values_i_2[i];
  }

  meas_values = (String)((int)(1 * (sum_data[0] / numOfSamples))) + "; " +
                (String)((int)(1 * (sum_data[1] / numOfSamples))) + "; " +
                (String)((int)(1 * (sum_data[2] / numOfSamples))) + "; " +
                (String)((int)(1 * (sum_data[3] / numOfSamples)));
  readRegRTD(1);
  meas_values = meas_values + "; " + (String)((1 * (convertToTemperature())));
  //LED_ch_x(1);
  readRegRTD(2);
  meas_values = meas_values + "; " + (String)((1 * (convertToTemperature())));
  //LED_ch_x(2);
  //delay(1);
  return (meas_values);
}
//----------------------------------------------------------------------------
void readADC()
{
  adcOutput res;
  //delay(1);
  while (!adc.isDataReady()) {}
  if (adc.isDataReady())
  {
    res = adc.readADC();
    //Serial.print("Status = ");
    //Serial.println(res.status, BIN);

    result = res.ch0;
    measuredDataADC[0] = (double)(1.2) / (double)(0x7FFFFF) * ((double)result) * ((double)(1.8 + 0.04)) * (U1_G_C) + (U1_O_C);
    //LED_ch_x(3);

    result = res.ch1;
    if (range_1 == 0)
    {
      measuredDataADC[1] = (double)(1.2) / (double)(0x7FFFFF) * ((double)result) * ((double)(1.0 / (80.0 * 0.01))) * (I1_UG_C) + (I1_UO_C);
    }
    else if (range_1 == 1)
    {
      measuredDataADC[1] = (double)(1.2) / (double)(0x7FFFFF) * ((double)result) * ((double)(1.0 / (80.0 * 0.11))) * (I1_LG_C) + (I1_LO_C);
    }
    //LED_ch_x(4);

    result = res.ch2;
    measuredDataADC[2] = (double)(1.2) / (double)(0x7FFFFF) * ((double)result) * ((double)(1.8 + 0.04)) * (U2_G_C) + (U2_O_C);
    //LED_ch_x(5);

    result = res.ch3;
    if (range_2 == 0)
    {
      measuredDataADC[3] = (double)(1.2) / (double)(0x7FFFFF) * ((double)result) * ((double)(1.0 / (80.0 * 0.01))) * (I2_UG_C) + (I2_UO_C);
    }
    else if (range_2 == 1)
    {
      measuredDataADC[3] = (double)(1.2) / (double)(0x7FFFFF) * ((double)result) * ((double)(1.0 / (80.0 * 0.11))) * (I2_LG_C) + (I2_LO_C);
    }
    //LED_ch_x(6);
    //delay(1);
  }
}
//----------------------------------------------------------------------------
void readRegRTD(int channelRTD)
{
  //delay(1);
  SPI.beginTransaction(SPISettings(50000, MSBFIRST, SPI_MODE1));

  if (channelRTD == 1)
  {
    digitalWrite(MAX1_CS, LOW);
  }
  else if (channelRTD == 2)
  {
    digitalWrite(MAX2_CS, LOW);
  }
  else {}
  SPI.transfer(0x80); //80h = 128 - config registr
  SPI.transfer(0xA3); //0xA3 -> 4W, 0xB3 -> 3W
  //delay(1);
  if (channelRTD == 1)
  {
    digitalWrite(MAX1_CS, HIGH);
  }
  else if (channelRTD == 2)
  {
    digitalWrite(MAX2_CS, HIGH);
  }
  else {}

  if (channelRTD == 1)
  {
    digitalWrite(MAX1_CS, LOW);
  }
  else if (channelRTD == 2)
  {
    digitalWrite(MAX2_CS, LOW);
  }
  else {}
  SPI.transfer(1);
  reg1 = SPI.transfer(0xFF);
  //SPI.transfer(0x02);
  reg2 = SPI.transfer(0xFF);
  if (channelRTD == 1)
  {
    digitalWrite(MAX1_CS, HIGH);
  }
  else if (channelRTD == 2)
  {
    digitalWrite(MAX2_CS, HIGH);
  }
  else {}

  fullreg = reg1;
  fullreg <<= 8;
  fullreg |= reg2;
  fullreg >>= 1; //D0 je chybovy bit, netvori merenou hodnotu
  resistance = fullreg; //predani jine promenne, ALE neni to jeste hodnota odporu RTD

  if (channelRTD == 1)
  {
    digitalWrite(MAX1_CS, LOW);
  }
  else if (channelRTD == 2)
  {
    digitalWrite(MAX2_CS, LOW);
  }
  else {}
  SPI.transfer(0x80); //80h = 128
  SPI.transfer(0x83); //0x83 -> 4W, 0x93 -> 3W
  SPI.endTransaction();
  if (channelRTD == 1)
  {
    digitalWrite(MAX1_CS, HIGH);
  }
  else if (channelRTD == 2)
  {
    digitalWrite(MAX2_CS, HIGH);
  }
  else {}
  //LED_ch_x(channelRTD);
  //delay(1);
}
//----------------------------------------------------------------------------
double convertToTemperature()
{
  Rt = resistance;
  Rt /= 32768.0;
  Rt *= 390.0; //jiz skutecna hodnota odporu RTD

  Z1 = -RTDa;
  Z2 = RTDa * RTDa - (4 * RTDb);
  Z3 = (4 * RTDb) / 100;
  Z4 = 2 * RTDb;

  //temperature = Z2 + (Z3 * Rt);
  //temperature = (sqrt(temperature) + Z1) / Z4;
  temperature = (-1.0 * 100.0 * RTDa + sqrt((100.0 * 100.0 * RTDa * RTDa) - (4 * 100.0 * RTDb * (100.0 - Rt)))) / (2.0 * 100.0 * RTDb);
  //temp = (long)(temperature * 10L);
  //temperature = (float)(temp/10.0);

  //Serial.print("Teplota: ");
  //Serial.print(temperature, 1);
  //Serial.println(" °C");

  if (temperature >= 0)
  {
    //Serial.print("Temperature: ");
    //Serial.println(temperature);
    return temperature; //exit
  }
  else
  {
    Rt /= 100;
    Rt *= 100;

    rpoly = Rt;

    temperature = -242.02;
    temperature += 2.2228 * rpoly;
    rpoly *= Rt; // square
    temperature += 2.5859e-3 * rpoly;
    rpoly *= Rt; // ^3
    temperature -= 4.8260e-6 * rpoly;
    rpoly *= Rt; // ^4
    temperature -= 2.8183e-8 * rpoly;
    rpoly *= Rt; // ^5
    temperature += 1.5243e-10 * rpoly;

    //Serial.print("Temperature: ");
    //Serial.println(temperature, 1);
    return temperature;
  }
  //Pozn: vsechny rovnice jsou uvedeny v AN-709  aplikačních poznámkách od AnalogDevices
}
//----------------------------------------------------------------------------
void blikLED(int pocet)
{
  for (int i = 0; i < pocet; i++)
  {
    uint8_t port_B;
    port_B = mcp.readPort(MCP23017Port::B);
    mcp.writePort(MCP23017Port::B, port_B & 0b01111111);
    flashTime = millis();
    while ((millis() - flashTime) < delay1) {}
    port_B = mcp.readPort(MCP23017Port::B);
    mcp.writePort(MCP23017Port::B, port_B | 0b10000000);
    flashTime = millis();
    while ((millis() - flashTime) < delay2) {}
  }
}
//----------------------------------------------------------------------------
void LED_ch_x(int kanal)
{
  if (kanal == 1)
  {
    mcp.writePort(MCP23017Port::A, 0b00111110);
  }
  else if (kanal == 2)
  {
    mcp.writePort(MCP23017Port::A, 0b00111101);
  }
  else if (kanal == 3)
  {
    mcp.writePort(MCP23017Port::A, 0b00111011);
  }
  else if (kanal == 4)
  {
    mcp.writePort(MCP23017Port::A, 0b00110111);
  }
  else if (kanal == 5)
  {
    mcp.writePort(MCP23017Port::A, 0b00101111);
  }
  else if (kanal == 6)
  {
    mcp.writePort(MCP23017Port::A, 0b00011111);
  }
  else {}
  flashTime = millis();
  while ((millis() - flashTime) < 8) {}
  mcp.writePort(MCP23017Port::A, 0b00111111);
  flashTime = millis();
  while ((millis() - flashTime) < 5) {}
}
//----------------------------------------------------------------------------
void KR()
{
  mcp.writePort(MCP23017Port::A, 0b00111110);
  flashTime = millis();
  while ((millis() - flashTime) < blueTime) {}
  mcp.writePort(MCP23017Port::A, 0b00111111);
  flashTime = millis();
  while ((millis() - flashTime) < blueTime) {}

  mcp.writePort(MCP23017Port::A, 0b00111101);
  flashTime = millis();
  while ((millis() - flashTime) < blueTime) {}
  mcp.writePort(MCP23017Port::A, 0b00111111);
  flashTime = millis();
  while ((millis() - flashTime) < blueTime) {}

  mcp.writePort(MCP23017Port::A, 0b00111011);
  flashTime = millis();
  while ((millis() - flashTime) < blueTime) {}
  mcp.writePort(MCP23017Port::A, 0b00111111);
  flashTime = millis();
  while ((millis() - flashTime) < blueTime) {}

  mcp.writePort(MCP23017Port::A, 0b00110111);
  flashTime = millis();
  while ((millis() - flashTime) < blueTime) {}
  mcp.writePort(MCP23017Port::A, 0b00111111);
  flashTime = millis();
  while ((millis() - flashTime) < blueTime) {}

  mcp.writePort(MCP23017Port::A, 0b00101111);
  flashTime = millis();
  while ((millis() - flashTime) < blueTime) {}
  mcp.writePort(MCP23017Port::A, 0b00111111);
  flashTime = millis();
  while ((millis() - flashTime) < blueTime) {}

  mcp.writePort(MCP23017Port::A, 0b00011111);
  flashTime = millis();
  while ((millis() - flashTime) < blueTime) {}
  mcp.writePort(MCP23017Port::A, 0b00111111);
  flashTime = millis();
  while ((millis() - flashTime) < blueTime) {}



  /*mcp.writePort(MCP23017Port::A, 0b00111111);
    flashTime = millis();
    while ((millis() - flashTime) < 100) {}
    mcp.writePort(MCP23017Port::A, 0b00111111);
    flashTime = millis();
    while ((millis() - flashTime) < 100) {}*/

  mcp.writePort(MCP23017Port::A, 0b00101111);
  flashTime = millis();
  while ((millis() - flashTime) < blueTime) {}
  mcp.writePort(MCP23017Port::A, 0b00111111);
  flashTime = millis();
  while ((millis() - flashTime) < blueTime) {}

  mcp.writePort(MCP23017Port::A, 0b00110111);
  flashTime = millis();
  while ((millis() - flashTime) < blueTime) {}
  mcp.writePort(MCP23017Port::A, 0b00111111);
  flashTime = millis();
  while ((millis() - flashTime) < blueTime) {}

  mcp.writePort(MCP23017Port::A, 0b00111011);
  flashTime = millis();
  while ((millis() - flashTime) < blueTime) {}
  mcp.writePort(MCP23017Port::A, 0b00111111);
  flashTime = millis();
  while ((millis() - flashTime) < blueTime) {}

  mcp.writePort(MCP23017Port::A, 0b00111101);
  flashTime = millis();
  while ((millis() - flashTime) < blueTime) {}
  mcp.writePort(MCP23017Port::A, 0b00111111);
  flashTime = millis();
  while ((millis() - flashTime) < blueTime) {}

  mcp.writePort(MCP23017Port::A, 0b00111110);
  flashTime = millis();
  while ((millis() - flashTime) < blueTime) {}
  mcp.writePort(MCP23017Port::A, 0b00111111);
  flashTime = millis();
  while ((millis() - flashTime) < blueTime) {}
}
//----------------------------------------------------------------------------
boolean checkRange()
{
  boolean stateOfRange = 0;
  uint8_t portA;
  readADC();

  if ((range_1 == 1) && (abs((measuredDataADC[1] * 1000)) > 125))
  {
    portA = mcp.readPort(MCP23017Port::A);
    mcp.writeRegister(MCP23017Register::GPIO_A, (portA & 0b10111111));
    range_1 = 0;
    stateOfRange = stateOfRange | 0;
    delayMicroseconds(100);
  }

  else if ((range_1 == 0) && (abs((measuredDataADC[1] * 1000)) < 120))
  {
    portA = mcp.readPort(MCP23017Port::A);
    mcp.writeRegister(MCP23017Register::GPIO_A, (portA | 0b01000000));
    range_1 = 1;
    stateOfRange = stateOfRange | 0;
    delayMicroseconds(100);
  }

  else if ((range_1 == 0) && (abs((measuredDataADC[1] * 1000)) > 1000))
  {
    if ((millis() - warningTime) > 500)
    {
      Serial.println("I1_OL");
      stateOfRange = stateOfRange | 1;
      warningTime = millis();
    }
  }

  if ((range_2 == 1) && (abs((measuredDataADC[3] * 1000)) > 125))
  {
    portA = mcp.readPort(MCP23017Port::A);
    mcp.writeRegister(MCP23017Register::GPIO_A, (portA & 0b01111111));
    range_2 = 0;
    stateOfRange = stateOfRange | 0;
    delayMicroseconds(100);
  }

  else if ((range_2 == 0) && (abs((measuredDataADC[3] * 1000)) < 120))
  {
    portA = mcp.readPort(MCP23017Port::A);
    mcp.writeRegister(MCP23017Register::GPIO_A, (portA | 0b10000000));
    range_2 = 1;
    stateOfRange = stateOfRange | 0;
    delayMicroseconds(100);
  }

  else if ((range_2 == 0) && (abs((measuredDataADC[3] * 1000)) > 1000))
  {
    if ((millis() - warningTime) > 500)
    {
      Serial.println("I2_OL");
      stateOfRange = stateOfRange | 1;
      warningTime = millis();
    }
  }
  //else{}

  if ((abs((measuredDataADC[0])) > 2.0))
  {
    if ((millis() - warningTime) > 500)
    {
      Serial.println("U1_OL");
      stateOfRange = stateOfRange | 1;
      warningTime = millis();
    }
  }

  if ((abs((measuredDataADC[2])) > 2.0))
  {
    if ((millis() - warningTime) > 500)
    {
      Serial.println("U2_OL");
      stateOfRange = stateOfRange | 1;
      warningTime = millis();
    }
  }

  return (stateOfRange);
}
//----------------------------------------------------------------------------
void setMCP()
{
  mcp.portMode(MCP23017Port::A, 0b00000000);//Port A as output
  mcp.portMode(MCP23017Port::B, 0b00000011);//Port B as out/in

  mcp.interruptMode(MCP23017InterruptMode::Separated);
  mcp.interrupt(MCP23017Port::B, FALLING);

  mcp.writeRegister(MCP23017Register::IPOL_A, 0x00);
  mcp.writeRegister(MCP23017Register::IPOL_B, 0x00);

  mcp.writeRegister(MCP23017Register::GPIO_A, 0b00111111);  //Reset port A
  mcp.writeRegister(MCP23017Register::GPIO_B, 0b11111100);  //Reset port B

  range_1 = 0;
  range_2 = 0;

  mcp.clearInterrupts();

  //blikLED(3);
}
//----------------------------------------------------------------------------
void setADC()
{
  adc.setInputChannelSelection(0, INPUT_CHANNEL_MUX_AIN0P_AIN0N);
  adc.setInputChannelSelection(1, INPUT_CHANNEL_MUX_AIN0P_AIN0N);
  adc.setInputChannelSelection(2, INPUT_CHANNEL_MUX_AIN0P_AIN0N);
  adc.setInputChannelSelection(3, INPUT_CHANNEL_MUX_AIN0P_AIN0N);
}
//----------------------------------------------------------------------------
void startSD(String filePath)
{
  if (digitalRead(SD_CD) == 0)
  {
    if (!SD.begin(SD_NCS, spiSD, 4000000)) {
      Serial.println("Card Mount Failed");
      return;
    }

    uint8_t cardType = SD.cardType();

    if (cardType == CARD_NONE) {
      Serial.println("No SD card attached");
      return;
    }

    Serial.print("SD Card Type: ");
    if (cardType == CARD_MMC) {
      Serial.println("MMC");
    } else if (cardType == CARD_SD) {
      Serial.println("SDSC");
    } else if (cardType == CARD_SDHC) {
      Serial.println("SDHC");
    } else {
      Serial.println("UNKNOWN");
    }

    uint64_t cardSize = SD.cardSize() / (1024 * 1024);
    Serial.printf("SD Card Size: % lluMB\n", cardSize);

    // If the data.csv file doesn't exist
    // Create a file on the SD card and write the data labels
    File file = SD.open(filePath);
    if (!file) {
      Serial.println("File doens't exist");
      Serial.println("Creating file...");
      writeFile(SD, filePath.c_str(), "Sample no.;U_1 (V);I_1 (A);U_2 (V);I_2 (A);T_1 (°C);T_2 (°C) \r\n");
    }
    else {
      Serial.println("File already exists");
    }
    file.close();
  }
  else
  {
    Serial.println("SD card not inserted");
  }
}
//----------------------------------------------------------------------------
void checkSD(String filePath)
{
  if (digitalRead(SD_CD) == 0)
  {
    // If the data.csv file doesn't exist
    // Create a file on the SD card and write the data labels
    File file = SD.open(filePath);
    if (!file) {
      Serial.println("File doens't exist");
      Serial.println("Creating file...");
      writeFile(SD, filePath.c_str(), "Sample no./Timestamp;U_1 (V);I_1 (A);U_2 (V);I_2 (A);T_1 (°C);T_2 (°C) \r\n");
    }
    else {
      Serial.println("File already exists");
    }
    file.close();
  }
  else
  {
    Serial.println("SD card not inserted");
  }
}

//--------------------------------------------------------------------------------------
//                    SD card default functions
//--------------------------------------------------------------------------------------
void listDir(fs::FS &fs, const char * dirname, uint8_t levels) {
  Serial.printf("Listing directory: % s\n", dirname);

  File root = fs.open(dirname);
  if (!root) {
    Serial.println("Failed to open directory");
    return;
  }
  if (!root.isDirectory()) {
    Serial.println("Not a directory");
    return;
  }

  File file = root.openNextFile();
  while (file) {
    if (file.isDirectory()) {
      Serial.print("  DIR : ");
      Serial.println(file.name());
      if (levels) {
        listDir(fs, file.name(), levels - 1);
      }
    } else {
      Serial.print("  FILE: ");
      Serial.print(file.name());
      Serial.print("  SIZE: ");
      Serial.println(file.size());
    }
    file = root.openNextFile();
  }
}
//----------------------------------------------------------------------------
void createDir(fs::FS &fs, const char * path) {
  Serial.printf("Creating Dir: % s\n", path);
  if (fs.mkdir(path)) {
    Serial.println("Dir created");
  } else {
    Serial.println("mkdir failed");
  }
}
//----------------------------------------------------------------------------
void removeDir(fs::FS &fs, const char * path) {
  Serial.printf("Removing Dir: % s\n", path);
  if (fs.rmdir(path)) {
    Serial.println("Dir removed");
  } else {
    Serial.println("rmdir failed");
  }
}
//----------------------------------------------------------------------------
void readFile(fs::FS &fs, const char * path) {
  Serial.printf("Reading file: % s\n", path);

  File file = fs.open(path);
  if (!file) {
    Serial.println("Failed to open file for reading");
    return;
  }

  Serial.print("Read from file : ");
  while (file.available()) {
    Serial.write(file.read());
  }
  file.close();
}
//----------------------------------------------------------------------------
void writeFile(fs::FS &fs, const char * path, const char * message) {
  Serial.printf("Writing file : % s\n", path);

  File file = fs.open(path, FILE_WRITE);
  if (!file) {
    Serial.println("Failed to open file for writing");
    return;
  }
  if (file.print(message)) {
    Serial.println("File written");
  } else {
    Serial.println("Write failed");
  }
  file.close();
}
//----------------------------------------------------------------------------
void appendFile(fs::FS &fs, const char * path, const char * message) {
  //Serial.printf("Appending to file : % s\n", path);

  File file = fs.open(path, FILE_APPEND);
  if (!file) {
    //Serial.println("Failed to open file for appending");
    return;
  }
  if (file.print(message)) {
    //Serial.println("Message appended");
    blikLED(1);
  } else {
    //Serial.println("Append failed");
  }
  file.close();
}
//----------------------------------------------------------------------------
void renameFile(fs::FS &fs, const char * path1, const char * path2) {
  Serial.printf("Renaming file % s to % s\n", path1, path2);
  if (fs.rename(path1, path2)) {
    Serial.println("File renamed");
  } else {
    Serial.println("Rename failed");
  }
}
//----------------------------------------------------------------------------
void deleteFile(fs::FS &fs, const char * path) {
  Serial.printf("Deleting file : % s\n", path);
  if (fs.remove(path)) {
    Serial.println("File deleted");
  } else {
    Serial.println("Delete failed");
  }
}
//----------------------------------------------------------------------------
void testFileIO(fs::FS &fs, const char * path) {
  File file = fs.open(path);
  static uint8_t buf[512];
  size_t len = 0;
  uint32_t start = millis();
  uint32_t end = start;
  if (file) {
    len = file.size();
    size_t flen = len;
    start = millis();
    while (len) {
      size_t toRead = len;
      if (toRead > 512) {
        toRead = 512;
      }
      file.read(buf, toRead);
      len -= toRead;
    }
    end = millis() - start;
    Serial.printf(" % u bytes read for % u ms\n", flen, end);
    file.close();
  } else {
    Serial.println("Failed to open file for reading");
  }


  file = fs.open(path, FILE_WRITE);
  if (!file) {
    Serial.println("Failed to open file for writing");
    return;
  }

  size_t i;
  start = millis();
  for (i = 0; i < 2048; i++) {
    file.write(buf, 512);
  }
  end = millis() - start;
  Serial.printf(" % u bytes written for % u ms\n", 2048 * 512, end);
  file.close();
}
